define(['jquery', 'underscore', 'backbone', 'tpls', 'views/content/query/filter/condition',
        'views/content/query/filter/sorting', 'squel', 'models/query/filter/condition',
        'collections/query/filter/conditions', 'collections/query/filter/sortings', 'models/query/filter/sorting',
        'events', 'context', 'baseview', 'collections/tables', 'views/content/query/history/list'],
    function ($, _, Backbone, Templates, ConditionView, SortingView, squel, Condition, Conditions, Sortings, Sorting,
              Events, Context, BaseView, Tables, HistoryList) {
        var FilterView = BaseView.extend({
            template: Templates["content/query/filter"],
            initialize: function () {

                // enable output panel, because it is not needed for this view
                Context.output = true;

                this.conditions = new Conditions();
                this.sortings = new Sortings();

                this.listenTo(this.sortings, 'reset', function(col, opts){
                  _.each(opts.previousModels, function(model) {
                    model.trigger('remove');
                  });
                });

                this.listenTo(this.conditions, 'add', this.renderCondition);
                this.listenTo(this.sortings, 'add', this.renderSorting);
                this.listenTo(Events, "table:change", this.setTable);
                this.listenTo(Events, "table:deleted", function () { this.updateTables(); }, this);
                this.listenTo(Events, "tables:fetched", this.updateTables, this);
                // after a table is created or renamed, context tables are
                // updated and tables:context-updated event triggered
                this.listenTo(Events, "tables:context-updated", this.updateTables, this);
                this.listenTo(Events, "db:change", this.setTable);
                this.listenTo(Events, "db:change", this.resetTables, this);
                this.listenTo(Events, "data:sort", this.sortData, this);

                _.bindAll(this, "resetViews", "resetTables");
            },
            events: {
                "click .add-condition": "addCondition",
                "click .add-sorting": "addSorting",
                "change select[name=table]": "changeTable",
                "submit form.filter-form": "executeQuery"
            },
            resetTables: function () {
                this.$("[name=table]").html($("<option>"));
            },
            changeTable: function (e) {
                var table = $(e.currentTarget).val();
                Events.trigger("table:navigate", table);
            },
            updateTables: function (tables, res) {
                if (!tables) {
                    // no tables passed, map and create collection from context
                    var mapped_tables = _.map(Context.tables, function (table) {
                        return {title: table};
                    });
                    tables = new Tables(mapped_tables);
                }

                this.$("[name=table]").html($("<option>"));
                tables.each(function (table) {
                    var $opt = $("<option>").text(table.get("title"));
                    if (Context.table === table.get("title")) {
                        $opt.prop("selected", true);
                    }
                    this.$("[name=table]").append($opt);
                });
            },
            setTable: function (table, meta) {

                // check if tables needs to be updated
                if (_.isEmpty(this.$("[name=table]").val())) {
                    // yep no tables, try to fill selectbox
                    this.updateTables();
                }

                var title, columns;

                if (_.isObject(table)) {
                    title = table.get("title");
                }

                if (_.isObject(meta)) {
                    columns = meta.getColumnString();
                }

                this.$("[name=table] option").prop("selected", false)
                    .filter(function () { return $(this).val() === title; })
                    .prop("selected", true);

                this.$("[name=columns]").val(columns);

                this.resetViews();
            },
            sortData: function (column, order) {
              this.sortings.reset();
              this.sortings.add(new Sorting({column: column, order: order}));
              Events.trigger("query:execute");
            },
            addCondition: function (e) {
                this.conditions.add(new Condition());
            },
            addSorting: function (e) {
                this.sortings.add(new Sorting());
            },
            resetViews: function () {
                this.conditions.reset();
                this.sortings.reset();
                _.each(this.subviews, function (view) {
                    view.remove();
                });

                // render history view into extra wrapper
                this.renderHistoryView();
            },
            renderCondition: function (condition) {
                this.subviews["condition-"+condition.cid] = new ConditionView({model: condition});
                this.$(".filter").append(this.subviews["condition-"+condition.cid].render().el);
                this.subviews["condition-"+condition.cid].focus();
            },
            renderSorting: function (sorting) {
                this.subviews["sorting-"+sorting.cid] = new SortingView({model: sorting});
                this.$(".filter").append(this.subviews["sorting-"+sorting.cid].render().el);
                this.subviews["sorting-"+sorting.cid].focus();
            },
            executeQuery: function (e) {
                e.preventDefault();

                // fill columns if the field is empty
                if (this.$("[name=columns]").val().trim().length === 0 && _.isArray(Context.meta.columns)) {
                    this.$("[name=columns]").val(Context.meta.columns.join(", "));
                }

                // trigger query:execute events
                Events.trigger("query:execute");
            },
            isReady: function () {
                return !_.isEmpty(this.$("[name=columns]").val()) && !_.isEmpty(this.$("[name=table]").val());
            },
            /**
             * Build base query for filter and raw query editor
             * @param table String
             * @returns squel
             */
            buildBaseQuery: function (table) {

                // query builder instance
                var s = squel.select(Context.getSquelSettings());

                // set columns
                _.each(Context.meta.columns, function (column) {
                    // escape all base query fields
                    s.field(Context.quote(column));
                });

                // set from table name
                s.from(table);

                // set base limit
                s.limit(20);

                return s;
            },
            getQuery: function () {

                // stop execution if filter is not yet ready
                if (!this.isReady()) { return null; }

                // build basic query as a starting point
                var s = squel.select(Context.getSquelSettings());
                s.from(this.$("[name=table]").val());

                // get all fields as array
                var fields = this.$("[name=columns]").val().split(",");

                if (this.$("[name=columns]").val().trim() === "*") {
                  // skip fields, as it will result in SELECT * FROM ...
                  // ...
                } else {
                  // trim each field and set it in query builder
                  _.each(fields, function (f) {
                      f = f.trim();
                      if (f.length > 0) {

                          // regex test if column matches function or alias pattern
                          // string is not a column name if it contains "(", ")" or " as ".
                          //var colTest = /[\(\)\s]/gi;
                          var colTest = /.*((\s(as)\s)|([\(\)])).*/gi;

                          // check if field is not function call or alias
                          // in that case we skip quoting
                          if (!colTest.test(f)) {
                              // regular field name, need to sanitize it
                              f = Context.quote(f);
                          }

                          // add field to query
                          s.field(f);

                      }
                  });
                }

                // set conditions
                this.conditions.each(function (condition) {
                    if (condition.get("column")) {
                        s.where(condition.queryString(), condition.get("value"));
                    }
                });

                // set sortings
                this.sortings.each(function (sorting) {
                    if (sorting.get("column").trim().length > 0) {
                        var order = sorting.get("order").toLowerCase() === "asc";
                        s.order(sorting.get("column"), order);
                    }
                });

                // set limit
                s.limit(this.$("[name=limit]").val() || 20);

                // set offset
                s.offset(this.$("[name=offset]").val() || 0);

                return s.toString();
            },
            render: function () {
                this.$el.html(this.template());

                // render history view into extra wrapper
                this.renderHistoryView();

                return this;
            },
            renderHistoryView: function () {
              // create history list view
              this.subviews.hv = new HistoryList();
              this.$(".extra-wrapper").html(this.subviews.hv.render().el);
            },
            escape: function (str) {
                return str.replace(/[\0\x08\x09\x1a\n\r"'\\\%]/g, function (char) {
                    switch (char) {
                        case "\0":
                            return "\\0";
                        case "\x08":
                            return "\\b";
                        case "\x09":
                            return "\\t";
                        case "\x1a":
                            return "\\z";
                        case "\n":
                            return "\\n";
                        case "\r":
                            return "\\r";
                        case "\"":
                        case "'":
                        case "\\":
                        case "%":
                            return "\\" + char;
                    }
                });
            }
        });
        return FilterView;
    });
